@@ -1,4 +1,6 @@
 /*
+ * SPDX-License-Identifier: ISC
+ *
  * Copyright (c) 2004-2005, 2010-2015, 2017-2018
  *	Todd C. Miller <Todd.Miller@sudo.ws>
  *
@@ -22,8 +24,6 @@
 
 #include <config.h>
 
-#include <sys/types.h>
-
 #include <errno.h>
 #include <netdb.h>
 #include <stdio.h>
@@ -32,22 +32,20 @@
 #ifdef HAVE_STDBOOL_H
 # include <stdbool.h>
 #else
-# include "compat/stdbool.h"
+# include <compat/stdbool.h>
 #endif /* HAVE_STDBOOL_H */
-
-#define DEFAULT_TEXT_DOMAIN	"sudo"
-#include "sudo_gettext.h"	/* must be included before sudo_compat.h */
-
-#include "sudo_compat.h"
-#include "sudo_fatal.h"
-#include "sudo_queue.h"
-#include "sudo_util.h"
-#include "sudo_plugin.h"
-
+#include <unistd.h>
 #ifndef HAVE_GETADDRINFO
-# include "compat/getaddrinfo.h"
+# include <compat/getaddrinfo.h>
 #endif
 
+#include <sudo_compat.h>
+#include <sudo_fatal.h>
+#include <sudo_gettext.h>
+#include <sudo_queue.h>
+#include <sudo_util.h>
+#include <sudo_plugin.h>
+
 struct sudo_fatal_callback {
     SLIST_ENTRY(sudo_fatal_callback) entries;
     void (*func)(void);
@@ -56,10 +54,10 @@ SLIST_HEAD(sudo_fatal_callback_list, sudo_fatal_callback);
 
 static struct sudo_fatal_callback_list callbacks = SLIST_HEAD_INITIALIZER(&callbacks);
 static sudo_conv_t sudo_warn_conversation;
-static bool (*sudo_warn_setlocale)(bool, int *);
-static bool (*sudo_warn_setlocale_prev)(bool, int *);
+static sudo_warn_setlocale_t sudo_warn_setlocale;
+static sudo_warn_setlocale_t sudo_warn_setlocale_prev;
 
-static void warning(const char *errstr, const char *fmt, va_list ap);
+static void warning(const char * restrict errstr, const char * restrict fmt, va_list ap);
 
 static void
 do_cleanup(void)
@@ -74,8 +72,8 @@ do_cleanup(void)
     }
 }
 
-void
-sudo_fatal_nodebug_v1(const char *fmt, ...)
+sudo_noreturn void
+sudo_fatal_nodebug_v1(const char * restrict fmt, ...)
 {
     va_list ap;
 
@@ -86,8 +84,8 @@ sudo_fatal_nodebug_v1(const char *fmt, ...)
     exit(EXIT_FAILURE);
 }
 
-void
-sudo_fatalx_nodebug_v1(const char *fmt, ...)
+sudo_noreturn void
+sudo_fatalx_nodebug_v1(const char * restrict fmt, ...)
 {
     va_list ap;
 
@@ -98,16 +96,16 @@ sudo_fatalx_nodebug_v1(const char *fmt, ...)
     exit(EXIT_FAILURE);
 }
 
-void
-sudo_vfatal_nodebug_v1(const char *fmt, va_list ap)
+sudo_noreturn void
+sudo_vfatal_nodebug_v1(const char * restrict fmt, va_list ap)
 {
     warning(strerror(errno), fmt, ap);
     do_cleanup();
     exit(EXIT_FAILURE);
 }
 
-void
-sudo_vfatalx_nodebug_v1(const char *fmt, va_list ap)
+sudo_noreturn void
+sudo_vfatalx_nodebug_v1(const char * restrict fmt, va_list ap)
 {
     warning(NULL, fmt, ap);
     do_cleanup();
@@ -115,7 +113,7 @@ sudo_vfatalx_nodebug_v1(const char *fmt, va_list ap)
 }
 
 void
-sudo_warn_nodebug_v1(const char *fmt, ...)
+sudo_warn_nodebug_v1(const char * restrict fmt, ...)
 {
     va_list ap;
 
@@ -125,7 +123,7 @@ sudo_warn_nodebug_v1(const char *fmt, ...)
 }
 
 void
-sudo_warnx_nodebug_v1(const char *fmt, ...)
+sudo_warnx_nodebug_v1(const char * restrict fmt, ...)
 {
     va_list ap;
     va_start(ap, fmt);
@@ -134,19 +132,19 @@ sudo_warnx_nodebug_v1(const char *fmt, ...)
 }
 
 void
-sudo_vwarn_nodebug_v1(const char *fmt, va_list ap)
+sudo_vwarn_nodebug_v1(const char * restrict fmt, va_list ap)
 {
     warning(strerror(errno), fmt, ap);
 }
 
 void
-sudo_vwarnx_nodebug_v1(const char *fmt, va_list ap)
+sudo_vwarnx_nodebug_v1(const char * restrict fmt, va_list ap)
 {
     warning(NULL, fmt, ap);
 }
 
-void
-sudo_gai_fatal_nodebug_v1(int errnum, const char *fmt, ...)
+sudo_noreturn void
+sudo_gai_fatal_nodebug_v1(int errnum, const char * restrict fmt, ...)
 {
     va_list ap;
 
@@ -157,8 +155,8 @@ sudo_gai_fatal_nodebug_v1(int errnum, const char *fmt, ...)
     exit(EXIT_FAILURE);
 }
 
-void
-sudo_gai_vfatal_nodebug_v1(int errnum, const char *fmt, va_list ap)
+sudo_noreturn void
+sudo_gai_vfatal_nodebug_v1(int errnum, const char * restrict fmt, va_list ap)
 {
     warning(gai_strerror(errnum), fmt, ap);
     do_cleanup();
@@ -166,7 +164,7 @@ sudo_gai_vfatal_nodebug_v1(int errnum, const char *fmt, va_list ap)
 }
 
 void
-sudo_gai_warn_nodebug_v1(int errnum, const char *fmt, ...)
+sudo_gai_warn_nodebug_v1(int errnum, const char * restrict fmt, ...)
 {
     va_list ap;
 
@@ -176,15 +174,16 @@ sudo_gai_warn_nodebug_v1(int errnum, const char *fmt, ...)
 }
 
 void
-sudo_gai_vwarn_nodebug_v1(int errnum, const char *fmt, va_list ap)
+sudo_gai_vwarn_nodebug_v1(int errnum, const char * restrict fmt, va_list ap)
 {
     warning(gai_strerror(errnum), fmt, ap);
 }
 
 static void
-warning(const char *errstr, const char *fmt, va_list ap)
+warning(const char * restrict errstr, const char * restrict fmt, va_list ap)
 {
     int cookie;
+    const int saved_errno = errno;
 
     /* Set user locale if setter was specified. */
     if (sudo_warn_setlocale != NULL)
@@ -206,17 +205,17 @@ warning(const char *errstr, const char *fmt, va_list ap)
 		va_copy(ap2, ap);
 		buflen = vsnprintf(static_buf, sizeof(static_buf), fmt, ap2);
 		va_end(ap2);
-		if (buflen >= (int)sizeof(static_buf)) {
-		    buf = malloc(++buflen);
-		    if (buf != NULL)
-			(void)vsnprintf(buf, buflen, fmt, ap);
-		    else
+		if (buflen >= ssizeof(static_buf)) {
+		    /* Not enough room in static buf, allocate dynamically. */
+		    if (vasprintf(&buf, fmt, ap) == -1)
 			buf = static_buf;
 		}
-		msgs[nmsgs].msg_type = SUDO_CONV_ERROR_MSG;
-		msgs[nmsgs++].msg = ": ";
-		msgs[nmsgs].msg_type = SUDO_CONV_ERROR_MSG;
-		msgs[nmsgs++].msg = buf;
+		if (buflen > 0) {
+		    msgs[nmsgs].msg_type = SUDO_CONV_ERROR_MSG;
+		    msgs[nmsgs++].msg = ": ";
+		    msgs[nmsgs].msg_type = SUDO_CONV_ERROR_MSG;
+		    msgs[nmsgs++].msg = buf;
+		}
         }
         if (errstr != NULL) {
 	    msgs[nmsgs].msg_type = SUDO_CONV_ERROR_MSG;
@@ -240,12 +239,19 @@ warning(const char *errstr, const char *fmt, va_list ap)
             fputs(": ", stderr);
             fputs(errstr, stderr);
         }
+#ifndef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
+        if (sudo_term_is_raw(fileno(stderr)))
+            putc('\r', stderr);
+#endif
         putc('\n', stderr);
     }
 
     /* Restore old locale as needed. */
     if (sudo_warn_setlocale != NULL)
 	sudo_warn_setlocale(true, &cookie);
+
+    /* Do not clobber errno. */
+    errno = saved_errno;
 }
 
 /*
@@ -278,18 +284,19 @@ sudo_fatal_callback_register_v1(sudo_fatal_callback_t func)
 int
 sudo_fatal_callback_deregister_v1(sudo_fatal_callback_t func)
 {
-    struct sudo_fatal_callback *cb, **prev;
+    struct sudo_fatal_callback *cb, *prev = NULL;
 
     /* Search for callback and remove if found, dupes are not allowed. */
-    SLIST_FOREACH_PREVPTR(cb, prev, &callbacks, entries) {
+    SLIST_FOREACH(cb, &callbacks, entries) {
 	if (cb->func == func) {
-	    if (cb == SLIST_FIRST(&callbacks))
+	    if (prev == NULL)
 		SLIST_REMOVE_HEAD(&callbacks, entries);
 	    else
-		SLIST_REMOVE_AFTER(*prev, entries);
+		SLIST_REMOVE_AFTER(prev, entries);
 	    free(cb);
 	    return 0;
 	}
+	prev = cb;
     }
 
     return -1;
@@ -310,7 +317,7 @@ sudo_warn_set_conversation_v1(sudo_conv_t conv)
  * locale for user warnings.
  */
 void
-sudo_warn_set_locale_func_v1(bool (*func)(bool, int *))
+sudo_warn_set_locale_func_v1(sudo_warn_setlocale_t func)
 {
     sudo_warn_setlocale_prev = sudo_warn_setlocale;
     sudo_warn_setlocale = func;
