@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2018 Todd C. Miller <Todd.Miller@sudo.ws>
+ * SPDX-License-Identifier: ISC
+ *
+ * Copyright (c) 2018-2023 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -21,23 +23,20 @@
 
 #include <config.h>
 
-#include <sys/types.h>
 #include <stdio.h>
 #include <stdlib.h>
-#ifdef HAVE_STRING_H
-# include <string.h>
-#endif /* HAVE_STRING_H */
+#include <string.h>
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif /* HAVE_STRINGS_H */
 #include <unistd.h>
 #include <stdarg.h>
 
-#include "sudoers.h"
-#include "sudo_ldap.h"
-#include "redblack.h"
-#include "cvtsudoers.h"
-#include "sudo_lbuf.h"
+#include <sudoers.h>
+#include <sudo_ldap.h>
+#include <redblack.h>
+#include <cvtsudoers.h>
+#include <sudo_lbuf.h>
 #include <gram.h>
 
 struct seen_user {
@@ -68,8 +67,9 @@ seen_user_free(void *v)
 static bool
 safe_string(const char *str)
 {
-    unsigned int ch = *str++;
-    debug_decl(safe_string, SUDOERS_DEBUG_UTIL)
+    const unsigned char *ustr = (const unsigned char *)str;
+    unsigned char ch = *ustr++;
+    debug_decl(safe_string, SUDOERS_DEBUG_UTIL);
 
     /* Initial char must be <= 127 and not LF, CR, SPACE, ':', '<' */
     switch (ch) {
@@ -87,7 +87,7 @@ safe_string(const char *str)
     }
 
     /* Any value <= 127 decimal except NUL, LF, and CR is safe */
-    while ((ch = *str++) != '\0') {
+    while ((ch = *ustr++) != '\0') {
 	if (ch > 127 || ch == '\n' || ch == '\r')
 	    debug_return_bool(false);
     }
@@ -101,7 +101,7 @@ print_attribute_ldif(FILE *fp, const char *name, const char *value)
     const unsigned char *uvalue = (unsigned char *)value;
     char *encoded = NULL;
     size_t esize;
-    debug_decl(print_attribute_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_attribute_ldif, SUDOERS_DEBUG_UTIL);
 
     if (!safe_string(value)) {
 	const size_t vlen = strlen(value);
@@ -127,12 +127,12 @@ print_attribute_ldif(FILE *fp, const char *name, const char *value)
  * Print sudoOptions from a defaults_list.
  */
 static bool
-print_options_ldif(FILE *fp, struct defaults_list *options)
+print_options_ldif(FILE *fp, const struct defaults_list *options)
 {
     struct defaults *opt;
     char *attr_val;
     int len;
-    debug_decl(print_options_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_options_ldif, SUDOERS_DEBUG_UTIL);
 
     TAILQ_FOREACH(opt, options, entries) {
 	if (opt->type != DEFAULTS)
@@ -162,14 +162,14 @@ print_options_ldif(FILE *fp, struct defaults_list *options)
  * Print global Defaults in a single sudoRole object.
  */
 static bool
-print_global_defaults_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
-    const char *base)
+print_global_defaults_ldif(FILE *fp,
+    const struct sudoers_parse_tree *parse_tree, const char *base)
 {
     unsigned int count = 0;
     struct sudo_lbuf lbuf;
     struct defaults *opt;
     char *dn;
-    debug_decl(print_global_defaults_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_global_defaults_ldif, SUDOERS_DEBUG_UTIL);
 
     sudo_lbuf_init(&lbuf, NULL, 0, NULL, 80);
 
@@ -181,8 +181,8 @@ print_global_defaults_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 	    lbuf.len = 0;
 	    sudo_lbuf_append(&lbuf, "# ");
 	    sudoers_format_default_line(&lbuf, parse_tree, opt, false, true);
-	    fprintf(fp, "# Unable to translate %s:%d\n%s\n",
-		opt->file, opt->lineno, lbuf.buf);
+	    fprintf(fp, "# Unable to translate %s:%d:%d:\n%s\n",
+		opt->file, opt->line, opt->column, lbuf.buf);
 	}
     }
     sudo_lbuf_destroy(&lbuf);
@@ -207,40 +207,81 @@ print_global_defaults_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
     debug_return_bool(!ferror(fp));
 }
 
+/*
+ * Format a sudo_command as a string.
+ * Returns the formatted, dynamically allocated string or dies on error.
+ */
+static char *
+format_cmnd(struct sudo_command *c, bool negated)
+{
+    struct command_digest *digest;
+    char *buf, *cp, *cmnd;
+    size_t bufsiz;
+    int len;
+    debug_decl(format_cmnd, SUDOERS_DEBUG_UTIL);
+
+    cmnd = c->cmnd ? c->cmnd : (char *)"ALL";
+    bufsiz = negated + strlen(cmnd) + 1;
+    if (c->args != NULL)
+	bufsiz += 1 + strlen(c->args);
+    TAILQ_FOREACH(digest, &c->digests, entries) {
+	bufsiz += strlen(digest_type_to_name(digest->digest_type)) + 1 +
+	    strlen(digest->digest_str) + 1;
+	if (TAILQ_NEXT(digest, entries) != NULL)
+	    bufsiz += 2;
+    }
+
+    if ((buf = malloc(bufsiz)) == NULL) {
+	sudo_fatalx(U_("%s: %s"), __func__,
+	    U_("unable to allocate memory"));
+    }
+
+    cp = buf;
+    TAILQ_FOREACH(digest, &c->digests, entries) {
+	len = snprintf(cp, bufsiz - (size_t)(cp - buf), "%s:%s%s ", 
+	    digest_type_to_name(digest->digest_type), digest->digest_str,
+	    TAILQ_NEXT(digest, entries) ? "," : "");
+	if (len < 0 || len >= (int)bufsiz - (cp - buf))
+	    sudo_fatalx(U_("internal error, %s overflow"), __func__);
+	cp += len;
+    }
+
+    len = snprintf(cp, bufsiz - (size_t)(cp - buf), "%s%s%s%s",
+	negated ? "!" : "", cmnd, c->args ? " " : "", c->args ? c->args : "");
+    if (len < 0 || len >= (int)bufsiz - (cp - buf))
+	sudo_fatalx(U_("internal error, %s overflow"), __func__);
+
+    debug_return_str(buf);
+}
+
 /*
  * Print struct member in LDIF format as the specified attribute.
  * See print_member_int() in parse.c.
  */
 static void
-print_member_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree, char *name,
-    int type, bool negated, int alias_type, const char *attr_name)
+print_member_ldif(FILE *fp, const struct sudoers_parse_tree *parse_tree,
+    char *name, int type, bool negated, short alias_type,
+    const char *attr_name)
 {
     struct alias *a;
     struct member *m;
-    struct sudo_command *c;
     char *attr_val;
     int len;
-    debug_decl(print_member_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_member_ldif, SUDOERS_DEBUG_UTIL);
 
     switch (type) {
-    case ALL:
-	print_attribute_ldif(fp, attr_name, negated ? "!ALL" : "ALL");
-	break;
     case MYSELF:
 	/* Only valid for sudoRunasUser */
 	print_attribute_ldif(fp, attr_name, "");
 	break;
-    case COMMAND:
-	c = (struct sudo_command *)name;
-	len = asprintf(&attr_val, "%s%s%s%s%s%s%s%s",
-	    c->digest ? digest_type_to_name(c->digest->digest_type) : "",
-	    c->digest ? ":" : "", c->digest ? c->digest->digest_str : "",
-	    c->digest ? " " : "", negated ? "!" : "", c->cmnd,
-	    c->args ? " " : "", c->args ? c->args : "");
-	if (len == -1) {
-	    sudo_fatalx(U_("%s: %s"), __func__,
-		U_("unable to allocate memory"));
+    case ALL:
+	if (name == NULL) {
+	    print_attribute_ldif(fp, attr_name, negated ? "!ALL" : "ALL");
+	    break;
 	}
+	FALLTHROUGH;
+    case COMMAND:
+	attr_val = format_cmnd((struct sudo_command *)name, negated);
 	print_attribute_ldif(fp, attr_name, attr_val);
 	free(attr_val);
 	break;
@@ -253,7 +294,7 @@ print_member_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree, char *name,
 	    alias_put(a);
 	    break;
 	}
-	/* FALLTHROUGH */
+	FALLTHROUGH;
     default:
 	len = asprintf(&attr_val, "%s%s", negated ? "!" : "", name);
 	if (len == -1) {
@@ -274,15 +315,17 @@ print_member_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree, char *name,
  * merge adjacent entries that are identical in all but the command.
  */
 static void
-print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
+print_cmndspec_ldif(FILE *fp, const struct sudoers_parse_tree *parse_tree,
     struct cmndspec *cs, struct cmndspec **nextp, struct defaults_list *options)
 {
+    char timebuf[sizeof("20120727121554Z")];
     struct cmndspec *next = *nextp;
     struct member *m;
-    struct tm *tp;
+    struct tm gmt;
+    char *attr_val;
     bool last_one;
-    char timebuf[sizeof("20120727121554Z")];
-    debug_decl(print_cmndspec_ldif, SUDOERS_DEBUG_UTIL)
+    size_t len;
+    debug_decl(print_cmndspec_ldif, SUDOERS_DEBUG_UTIL);
 
     /* Print runasuserlist as sudoRunAsUser attributes */
     if (cs->runasuserlist != NULL) {
@@ -302,22 +345,26 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 
     /* Print sudoNotBefore and sudoNotAfter attributes */
     if (cs->notbefore != UNSPEC) {
-	if ((tp = gmtime(&cs->notbefore)) == NULL) {
-	    sudo_warn(U_("unable to get GMT time"));
+	if (gmtime_r(&cs->notbefore, &gmt) == NULL) {
+	    sudo_warn("%s", U_("unable to get GMT time"));
 	} else {
-	    if (strftime(timebuf, sizeof(timebuf), "%Y%m%d%H%M%SZ", tp) == 0) {
-		sudo_warnx(U_("unable to format timestamp"));
+	    timebuf[sizeof(timebuf) - 1] = '\0';
+	    len = strftime(timebuf, sizeof(timebuf), "%Y%m%d%H%M%SZ", &gmt);
+	    if (len == 0 || timebuf[sizeof(timebuf) - 1] != '\0') {
+		sudo_warnx("%s", U_("unable to format timestamp"));
 	    } else {
 		print_attribute_ldif(fp, "sudoNotBefore", timebuf);
 	    }
 	}
     }
     if (cs->notafter != UNSPEC) {
-	if ((tp = gmtime(&cs->notafter)) == NULL) {
-	    sudo_warn(U_("unable to get GMT time"));
+	if (gmtime_r(&cs->notafter, &gmt) == NULL) {
+	    sudo_warn("%s", U_("unable to get GMT time"));
 	} else {
-	    if (strftime(timebuf, sizeof(timebuf), "%Y%m%d%H%M%SZ", tp) == 0) {
-		sudo_warnx(U_("unable to format timestamp"));
+	    timebuf[sizeof(timebuf) - 1] = '\0';
+	    len = strftime(timebuf, sizeof(timebuf), "%Y%m%d%H%M%SZ", &gmt);
+	    if (len == 0 || timebuf[sizeof(timebuf) - 1] != '\0') {
+		sudo_warnx("%s", U_("unable to format timestamp"));
 	    } else {
 		print_attribute_ldif(fp, "sudoNotAfter", timebuf);
 	    }
@@ -326,7 +373,6 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 
     /* Print timeout as a sudoOption. */
     if (cs->timeout > 0) {
-	char *attr_val;
 	if (asprintf(&attr_val, "command_timeout=%d", cs->timeout) == -1) {
 	    sudo_fatalx(U_("%s: %s"), __func__,
 		U_("unable to allocate memory"));
@@ -347,6 +393,10 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 	    print_attribute_ldif(fp, "sudoOption",
 		tag.noexec ? "noexec" : "!noexec");
 	}
+	if (tag.intercept != UNSPEC) {
+	    print_attribute_ldif(fp, "sudoOption",
+		tag.intercept ? "intercept" : "!intercept");
+	}
 	if (tag.send_mail != UNSPEC) {
 	    if (tag.send_mail) {
 		print_attribute_ldif(fp, "sudoOption", "mail_all_cmnds");
@@ -375,22 +425,35 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
     }
     print_options_ldif(fp, options);
 
+    /* Print runchroot and runcwd. */
+    if (cs->runchroot != NULL) {
+	if (asprintf(&attr_val, "runchroot=%s", cs->runchroot) == -1) {
+	    sudo_fatalx(U_("%s: %s"), __func__,
+		U_("unable to allocate memory"));
+	}
+	print_attribute_ldif(fp, "sudoOption", attr_val);
+	free(attr_val);
+    }
+    if (cs->runcwd != NULL) {
+	if (asprintf(&attr_val, "runcwd=%s", cs->runcwd) == -1) {
+	    sudo_fatalx(U_("%s: %s"), __func__,
+		U_("unable to allocate memory"));
+	}
+	print_attribute_ldif(fp, "sudoOption", attr_val);
+	free(attr_val);
+    }
+
 #ifdef HAVE_SELINUX
     /* Print SELinux role/type */
     if (cs->role != NULL && cs->type != NULL) {
-	char *attr_val;
-	int len;
-
-	len = asprintf(&attr_val, "role=%s", cs->role);
-	if (len == -1) {
+	if (asprintf(&attr_val, "role=%s", cs->role) == -1) {
 	    sudo_fatalx(U_("%s: %s"), __func__,
 		U_("unable to allocate memory"));
 	}
 	print_attribute_ldif(fp, "sudoOption", attr_val);
 	free(attr_val);
 
-	len = asprintf(&attr_val, "type=%s", cs->type);
-	if (len == -1) {
+	if (asprintf(&attr_val, "type=%s", cs->type) == -1) {
 	    sudo_fatalx(U_("%s: %s"), __func__,
 		U_("unable to allocate memory"));
 	}
@@ -399,15 +462,23 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
     }
 #endif /* HAVE_SELINUX */
 
+#ifdef HAVE_APPARMOR
+    /* Print AppArmor profile */
+    if (cs->apparmor_profile != NULL) {
+	if (asprintf(&attr_val, "apparmor_profile=%s", cs->apparmor_profile) == -1) {
+	    sudo_fatalx(U_("%s: %s"), __func__,
+		U_("unable to allocate memory"));
+	}
+	print_attribute_ldif(fp, "sudoOption", attr_val);
+	free(attr_val);
+    }
+#endif /* HAVE_APPARMOR */
+
 #ifdef HAVE_PRIV_SET
     /* Print Solaris privs/limitprivs */
     if (cs->privs != NULL || cs->limitprivs != NULL) {
-	char *attr_val;
-	int len;
-
 	if (cs->privs != NULL) {
-	    len = asprintf(&attr_val, "privs=%s", cs->privs);
-	    if (len == -1) {
+	    if (asprintf(&attr_val, "privs=%s", cs->privs) == -1) {
 		sudo_fatalx(U_("%s: %s"), __func__,
 		    U_("unable to allocate memory"));
 	    }
@@ -415,8 +486,7 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 	    free(attr_val);
 	}
 	if (cs->limitprivs != NULL) {
-	    len = asprintf(&attr_val, "limitprivs=%s", cs->limitprivs);
-	    if (len == -1) {
+	    if (asprintf(&attr_val, "limitprivs=%s", cs->limitprivs) == -1) {
 		sudo_fatalx(U_("%s: %s"), __func__,
 		    U_("unable to allocate memory"));
 	    }
@@ -442,7 +512,7 @@ print_cmndspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 #ifdef HAVE_SELINUX
 	    || cs->role != next->role || cs->type != next->type
 #endif /* HAVE_SELINUX */
-	    ;
+	    || cs->runchroot != next->runchroot || cs->runcwd != next->runcwd;
 
 	print_member_ldif(fp, parse_tree, cs->cmnd->name, cs->cmnd->type,
 	    cs->cmnd->negated, CMNDALIAS, "sudoCommand");
@@ -469,7 +539,7 @@ user_to_cn(const char *user)
     const char *src;
     char *cn, *dst;
     size_t size;
-    debug_decl(user_to_cn, SUDOERS_DEBUG_UTIL)
+    debug_decl(user_to_cn, SUDOERS_DEBUG_UTIL);
 
     /* Allocate as much as we could possibly need. */
     size = (2 * strlen(user)) + 64 + 1;
@@ -539,13 +609,13 @@ bad:
  * Print a single User_Spec.
  */
 static bool
-print_userspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
+print_userspec_ldif(FILE *fp, const struct sudoers_parse_tree *parse_tree,
     struct userspec *us, struct cvtsudoers_config *conf)
 {
     struct privilege *priv;
     struct member *m;
     struct cmndspec *cs, *next;
-    debug_decl(print_userspec_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_userspec_ldif, SUDOERS_DEBUG_UTIL);
 
     /*
      * Each userspec struct may contain multiple privileges for
@@ -588,7 +658,7 @@ print_userspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
 	    print_cmndspec_ldif(fp, parse_tree, cs, &next, &priv->defaults);
 
 	    if (conf->sudo_order != 0) {
-		char numbuf[(((sizeof(conf->sudo_order) * 8) + 2) / 3) + 2];
+		char numbuf[STRLEN_MAX_UNSIGNED(conf->sudo_order) + 1];
 		if (conf->order_max != 0 && conf->sudo_order > conf->order_max) {
 		    sudo_fatalx(U_("too many sudoers entries, maximum %u"),
 			conf->order_padding);
@@ -608,11 +678,11 @@ print_userspec_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
  * Print User_Specs.
  */
 static bool
-print_userspecs_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
+print_userspecs_ldif(FILE *fp, const struct sudoers_parse_tree *parse_tree,
     struct cvtsudoers_config *conf)
 {
     struct userspec *us;
-    debug_decl(print_userspecs_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(print_userspecs_ldif, SUDOERS_DEBUG_UTIL);
  
     TAILQ_FOREACH(us, &parse_tree->userspecs, entries) {
 	if (!print_userspec_ldif(fp, parse_tree, us, conf))
@@ -625,15 +695,15 @@ print_userspecs_ldif(FILE *fp, struct sudoers_parse_tree *parse_tree,
  * Export the parsed sudoers file in LDIF format.
  */
 bool
-convert_sudoers_ldif(struct sudoers_parse_tree *parse_tree,
+convert_sudoers_ldif(const struct sudoers_parse_tree *parse_tree,
     const char *output_file, struct cvtsudoers_config *conf)
 {
     bool ret = true;
     FILE *output_fp = stdout;
-    debug_decl(convert_sudoers_ldif, SUDOERS_DEBUG_UTIL)
+    debug_decl(convert_sudoers_ldif, SUDOERS_DEBUG_UTIL);
 
     if (conf->sudoers_base == NULL) {
-	sudo_fatalx(U_("the SUDOERS_BASE environment variable is not set and the -b option was not specified."));
+	sudo_fatalx("%s", U_("the SUDOERS_BASE environment variable is not set and the -b option was not specified."));
     }
 
     if (output_file != NULL && strcmp(output_file, "-") != 0) {
