@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2009-2017 Todd C. Miller <Todd.Miller@sudo.ws>
+ * SPDX-License-Identifier: ISC
+ *
+ * Copyright (c) 2009-2023 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -21,17 +23,10 @@
 
 #include <config.h>
 
-#include <sys/types.h>
 #include <sys/resource.h>
 #include <sys/stat.h>
-#include <stdio.h>
 #include <stdlib.h>
-#ifdef HAVE_STRING_H
-# include <string.h>
-#endif /* HAVE_STRING_H */
-#ifdef HAVE_STRINGS_H
-# include <strings.h>
-#endif /* HAVE_STRINGS_H */
+#include <string.h>
 #include <unistd.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -48,54 +43,44 @@
 # include <sys/task.h>
 #endif
 
-#include "sudo.h"
-#include "sudo_exec.h"
-#include "sudo_event.h"
-#include "sudo_plugin.h"
-#include "sudo_plugin_int.h"
+#include <sudo.h>
+#include <sudo_exec.h>
+#include <sudo_plugin.h>
+#include <sudo_plugin_int.h>
 
-#ifdef __linux__
-static struct rlimit nproclimit;
-#endif
-
-/*
- * Unlimit the number of processes since Linux's setuid() will
- * apply resource limits when changing uid and return EAGAIN if
- * nproc would be exceeded by the uid switch.
- */
+#ifdef HAVE_PTRACE_INTERCEPT
 static void
-unlimit_nproc(void)
+handler(int signo)
 {
-#ifdef __linux__
-    struct rlimit rl;
-    debug_decl(unlimit_nproc, SUDO_DEBUG_UTIL)
-
-    if (getrlimit(RLIMIT_NPROC, &nproclimit) != 0)
-	sudo_warn("getrlimit");
-    rl.rlim_cur = rl.rlim_max = RLIM_INFINITY;
-    if (setrlimit(RLIMIT_NPROC, &rl) != 0) {
-	rl.rlim_cur = rl.rlim_max = nproclimit.rlim_max;
-	if (setrlimit(RLIMIT_NPROC, &rl) != 0)
-	    sudo_warn("setrlimit");
-    }
-    debug_return;
-#endif /* __linux__ */
+    /* just return */
 }
+#endif /* HAVE_PTRACE_INTERCEPT */
 
-/*
- * Restore saved value of RLIMIT_NPROC.
- */
 static void
-restore_nproc(void)
+close_fds(struct command_details *details, int errfd, int intercept_fd)
 {
-#ifdef __linux__
-    debug_decl(restore_nproc, SUDO_DEBUG_UTIL)
+    int fd, maxfd;
+    unsigned char *debug_fds;
+    debug_decl(close_fds, SUDO_DEBUG_EXEC);
 
-    if (setrlimit(RLIMIT_NPROC, &nproclimit) != 0)
-	sudo_warn("setrlimit");
+    if (details->closefrom < 0)
+	debug_return;
+
+    /* Preserve debug fds and error pipe as needed. */
+    maxfd = sudo_debug_get_fds(&debug_fds);
+    for (fd = 0; fd <= maxfd; fd++) {
+	if (sudo_isset(debug_fds, fd))
+	    add_preserved_fd(&details->preserved_fds, fd);
+    }
+    if (errfd != -1)
+	add_preserved_fd(&details->preserved_fds, errfd);
+    if (intercept_fd != -1)
+	add_preserved_fd(&details->preserved_fds, intercept_fd);
+
+    /* Close all fds except those explicitly preserved. */
+    closefrom_except(details->closefrom, &details->preserved_fds);
 
     debug_return;
-#endif /* __linux__ */
 }
 
 /*
@@ -104,22 +89,17 @@ restore_nproc(void)
  * Returns true on success and false on failure.
  */
 static bool
-exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
+exec_setup(struct command_details *details, int intercept_fd, int errfd)
 {
     bool ret = false;
-    debug_decl(exec_setup, SUDO_DEBUG_EXEC)
+    debug_decl(exec_setup, SUDO_DEBUG_EXEC);
 
-#ifdef HAVE_SELINUX
-    if (ISSET(details->flags, CD_RBAC_ENABLED)) {
-	if (selinux_setup(details->selinux_role, details->selinux_type,
-	    ptyname ? ptyname : user_details.tty, ptyfd) == -1)
+#ifdef HAVE_PTRACE_INTERCEPT
+    if (ISSET(details->flags, CD_USE_PTRACE)) {
+	if (!set_exec_filter())
 	    goto done;
     }
-#endif
-
-    /* Restore coredumpsize resource limit before running. */
-    if (sudo_conf_disable_coredump())
-	disable_coredump(true);
+#endif /* HAVE_PTRACE_INTERCEPT */
 
     if (details->pw != NULL) {
 #ifdef HAVE_PROJECT_H
@@ -128,37 +108,37 @@ exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
 #ifdef HAVE_PRIV_SET
 	if (details->privs != NULL) {
 	    if (setppriv(PRIV_SET, PRIV_INHERITABLE, details->privs) != 0) {
-		sudo_warn("unable to set privileges");
+		sudo_warn("%s", U_("unable to set privileges"));
 		goto done;
 	    }
 	}
 	if (details->limitprivs != NULL) {
 	    if (setppriv(PRIV_SET, PRIV_LIMIT, details->limitprivs) != 0) {
-		sudo_warn("unable to set limit privileges");
+		sudo_warn("%s", U_("unable to set limit privileges"));
 		goto done;
 	    }
 	} else if (details->privs != NULL) {
 	    if (setppriv(PRIV_SET, PRIV_LIMIT, details->privs) != 0) {
-		sudo_warn("unable to set limit privileges");
+		sudo_warn("%s", U_("unable to set limit privileges"));
 		goto done;
 	    }
 	}
 #endif /* HAVE_PRIV_SET */
 
 #ifdef HAVE_GETUSERATTR
-	if (aix_prep_user(details->pw->pw_name, ptyname ? ptyname : user_details.tty) != 0) {
+	if (aix_prep_user(details->pw->pw_name, details->tty) != 0) {
 	    /* error message displayed by aix_prep_user */
 	    goto done;
 	}
 #endif
 #ifdef HAVE_LOGIN_CAP_H
 	if (details->login_class) {
-	    int flags;
+	    unsigned int flags;
 	    login_cap_t *lc;
 
 	    /*
-	     * We only use setusercontext() to set the nice value and rlimits
-	     * unless this is a login shell (sudo -i).
+	     * We only use setusercontext() to set the nice value, rlimits
+	     * and umask unless this is a login shell (sudo -i).
 	     */
 	    lc = login_getclass((char *)details->login_class);
 	    if (!lc) {
@@ -170,12 +150,11 @@ exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
 		/* Set everything except user, group and login name. */
 		flags = LOGIN_SETALL;
 		CLR(flags, LOGIN_SETGROUP|LOGIN_SETLOGIN|LOGIN_SETUSER|LOGIN_SETENV|LOGIN_SETPATH);
-		CLR(details->flags, CD_SET_UMASK); /* LOGIN_UMASK instead */
 	    } else {
-		flags = LOGIN_SETRESOURCES|LOGIN_SETPRIORITY;
+		flags = LOGIN_SETRESOURCES|LOGIN_SETPRIORITY|LOGIN_SETUMASK;
 	    }
 	    if (setusercontext(lc, details->pw, details->pw->pw_uid, flags)) {
-		sudo_warn(U_("unable to set user context"));
+		sudo_warn("%s", U_("unable to set user context"));
 		if (details->pw->pw_uid != ROOT_UID)
 		    goto done;
 	    }
@@ -191,12 +170,21 @@ exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
 
     if (ISSET(details->flags, CD_SET_PRIORITY)) {
 	if (setpriority(PRIO_PROCESS, 0, details->priority) != 0) {
-	    sudo_warn(U_("unable to set process priority"));
+	    sudo_warn("%s", U_("unable to set process priority"));
 	    goto done;
 	}
     }
-    if (ISSET(details->flags, CD_SET_UMASK))
+
+    /* Policy may override umask in PAM or login.conf. */
+    if (ISSET(details->flags, CD_OVERRIDE_UMASK))
 	(void) umask(details->umask);
+
+    /* Apply resource limits specified by the policy, if any. */
+    set_policy_rlimits();
+
+    /* Close fds before chroot (need /dev) or uid change (prlimit on Linux). */
+    close_fds(details, errfd, intercept_fd);
+
     if (details->chroot) {
 	if (chroot(details->chroot) != 0 || chdir("/") != 0) {
 	    sudo_warn(U_("unable to change root to %s"), details->chroot);
@@ -204,29 +192,29 @@ exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
 	}
     }
 
-    /* 
+    /*
      * Unlimit the number of processes since Linux's setuid() will
      * return EAGAIN if RLIMIT_NPROC would be exceeded by the uid switch.
      */
     unlimit_nproc();
 
 #if defined(HAVE_SETRESUID)
-    if (setresuid(details->uid, details->euid, details->euid) != 0) {
+    if (setresuid(details->cred.uid, details->cred.euid, details->cred.euid) != 0) {
 	sudo_warn(U_("unable to change to runas uid (%u, %u)"),
-	    (unsigned int)details->uid, (unsigned int)details->euid);
+	    (unsigned int)details->cred.uid, (unsigned int)details->cred.euid);
 	goto done;
     }
 #elif defined(HAVE_SETREUID)
-    if (setreuid(details->uid, details->euid) != 0) {
+    if (setreuid(details->cred.uid, details->cred.euid) != 0) {
 	sudo_warn(U_("unable to change to runas uid (%u, %u)"),
-	    (unsigned int)details->uid, (unsigned int)details->euid);
+	    (unsigned int)details->cred.uid, (unsigned int)details->cred.euid);
 	goto done;
     }
 #else
-    /* Cannot support real user ID that is different from effective user ID. */
-    if (setuid(details->euid) != 0) {
+    /* Cannot support real user-ID that is different from effective user-ID. */
+    if (setuid(details->cred.euid) != 0) {
 	sudo_warn(U_("unable to change to runas uid (%u, %u)"),
-	    (unsigned int)details->euid, (unsigned int)details->euid);
+	    (unsigned int)details->cred.euid, (unsigned int)details->cred.euid);
 	goto done;
     }
 #endif /* !HAVE_SETRESUID && !HAVE_SETREUID */
@@ -238,13 +226,22 @@ exec_setup(struct command_details *details, const char *ptyname, int ptyfd)
      * Only change cwd if we have chroot()ed or the policy modules
      * specifies a different cwd.  Must be done after uid change.
      */
-    if (details->cwd != NULL) {
-	if (details->chroot || user_details.cwd == NULL ||
-	    strcmp(details->cwd, user_details.cwd) != 0) {
-	    /* Note: cwd is relative to the new root, if any. */
-	    if (chdir(details->cwd) != 0) {
-		sudo_warn(U_("unable to change directory to %s"), details->cwd);
-		goto done;
+    if (details->runcwd != NULL) {
+	if (details->chroot != NULL || details->submitcwd == NULL ||
+	    strcmp(details->runcwd, details->submitcwd) != 0) {
+	    if (ISSET(details->flags, CD_RBAC_ENABLED)) {
+		 /* For SELinux, chdir(2) in sesh after the context change. */
+		SET(details->flags, CD_RBAC_SET_CWD);
+	    } else {
+		/* Note: runcwd is relative to the new root, if any. */
+		if (chdir(details->runcwd) == -1) {
+		    sudo_warn(U_("unable to change directory to %s"),
+			details->runcwd);
+		    if (!ISSET(details->flags, CD_CWD_OPTIONAL))
+			goto done;
+		    if (details->chroot != NULL)
+			sudo_warnx(U_("starting from %s"), "/");
+		}
 	    }
 	}
     }
@@ -262,38 +259,47 @@ done:
  * If the exec fails, cstat is filled in with the value of errno.
  */
 void
-exec_cmnd(struct command_details *details, int errfd)
+exec_cmnd(struct command_details *details, sigset_t *mask,
+    int intercept_fd, int errfd)
 {
-    debug_decl(exec_cmnd, SUDO_DEBUG_EXEC)
+    debug_decl(exec_cmnd, SUDO_DEBUG_EXEC);
 
-    restore_signals();
-    if (exec_setup(details, NULL, -1) == true) {
-	/* headed for execve() */
-	if (details->closefrom >= 0) {
-	    int fd, maxfd;
-	    unsigned char *debug_fds;
+#ifdef HAVE_PTRACE_INTERCEPT
+    if (ISSET(details->flags, CD_USE_PTRACE)) {
+	struct sigaction sa;
+	sigset_t set;
 
-	    /* Preserve debug fds and error pipe as needed. */
-	    maxfd = sudo_debug_get_fds(&debug_fds);
-	    for (fd = 0; fd <= maxfd; fd++) {
-		if (sudo_isset(debug_fds, fd))
-		    add_preserved_fd(&details->preserved_fds, fd);
-	    }
-	    if (errfd != -1)
-		add_preserved_fd(&details->preserved_fds, errfd);
-
-	    /* Close all fds except those explicitly preserved. */
-	    closefrom_except(details->closefrom, &details->preserved_fds);
+	/* Tracer will send us SIGUSR1 when it is time to proceed. */
+	memset(&sa, 0, sizeof(sa));
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART;
+	sa.sa_handler = handler;
+	if (sudo_sigaction(SIGUSR1, &sa, NULL) != 0) {
+	    sudo_warn(U_("unable to set handler for signal %d"),
+		SIGUSR1);
 	}
+
+	/* Suspend child until tracer seizes control and sends SIGUSR1. */
+	sigfillset(&set);
+	sigdelset(&set, SIGUSR1);
+	sigsuspend(&set);
+    }
+#endif /* HAVE_PTRACE_INTERCEPT */
+
+    if (mask != NULL)
+	sigprocmask(SIG_SETMASK, mask, NULL);
+    restore_signals();
+    if (exec_setup(details, intercept_fd, errfd) == true) {
+	/* headed for execve() */
 #ifdef HAVE_SELINUX
 	if (ISSET(details->flags, CD_RBAC_ENABLED)) {
 	    selinux_execve(details->execfd, details->command, details->argv,
-		details->envp, ISSET(details->flags, CD_NOEXEC));
+		details->envp, details->runcwd, details->flags);
 	} else
 #endif
 	{
 	    sudo_execve(details->execfd, details->command, details->argv,
-		details->envp, ISSET(details->flags, CD_NOEXEC));
+		details->envp, intercept_fd, details->flags);
 	}
     }
     sudo_debug_printf(SUDO_DEBUG_ERROR, "unable to exec %s: %s",
@@ -311,7 +317,7 @@ sudo_terminated(struct command_status *cstat)
 {
     int signo;
     bool sigtstp = false;
-    debug_decl(sudo_terminated, SUDO_DEBUG_EXEC)
+    debug_decl(sudo_terminated, SUDO_DEBUG_EXEC);
 
     for (signo = 0; signo < NSIG; signo++) {
 	if (signal_pending(signo)) {
@@ -354,11 +360,8 @@ sudo_terminated(struct command_status *cstat)
     debug_return_bool(false);
 }
 
-#if SUDO_API_VERSION != SUDO_API_MKVERSION(1, 13)
-# error "Update sudo_needs_pty() after changing the plugin API"
-#endif
 static bool
-sudo_needs_pty(struct command_details *details)
+sudo_needs_pty(const struct command_details *details)
 {
     struct plugin_container *plugin;
 
@@ -367,27 +370,90 @@ sudo_needs_pty(struct command_details *details)
 
     TAILQ_FOREACH(plugin, &io_plugins, entries) {
 	if (plugin->u.io->log_ttyin != NULL ||
-	    plugin->u.io->log_ttyout != NULL ||
-	    plugin->u.io->log_stdin != NULL ||
-	    plugin->u.io->log_stdout != NULL ||
-	    plugin->u.io->log_stderr != NULL ||
-	    plugin->u.io->change_winsize != NULL ||
-	    plugin->u.io->log_suspend != NULL)
+	    plugin->u.io->log_ttyout != NULL)
 	    return true;
     }
     return false;
 }
 
 /*
- * Execute a command, potentially in a pty with I/O loggging, and
+ * Check whether the specified fd matches the device file that
+ * corresponds to tty_sb.  If tty_sb is NULL, just check whether
+ * fd is a tty.  Always fills in fd_sb (zeroed on error).
+ * Returns true on match, else false.
+ */
+bool
+fd_matches_tty(int fd, struct stat *tty_sb, struct stat *fd_sb)
+{
+    debug_decl(fd_is_user_tty, SUDO_DEBUG_EXEC);
+
+    if (fstat(fd, fd_sb) == -1) {
+	 /* Always initialize fd_sb. */
+	memset(fd_sb, 0, sizeof(*fd_sb));
+	debug_return_bool(false);
+    }
+    if (!S_ISCHR(fd_sb->st_mode))
+	debug_return_bool(false);
+
+    /* Compare with tty_sb if available, else just check that fd is a tty. */
+    debug_return_bool(tty_sb ? tty_sb->st_rdev == fd_sb->st_rdev : isatty(fd));
+}
+
+/*
+ * If we are not running the command in a pty, we were not invoked as
+ * sudoedit, there is no command timeout and there is no close function,
+ * sudo can exec the command directly (and not wait).
+ */
+static bool
+direct_exec_allowed(const struct command_details *details)
+{
+    struct plugin_container *plugin;
+    debug_decl(direct_exec_allowed, SUDO_DEBUG_EXEC);
+
+    /* Assumes sudo_needs_pty() was already checked. */
+    if (ISSET(details->flags, CD_RBAC_ENABLED|CD_SET_TIMEOUT|CD_SUDOEDIT) ||
+	    policy_plugin.u.policy->close != NULL)
+	debug_return_bool(false);
+
+    TAILQ_FOREACH(plugin, &audit_plugins, entries) {
+	if (plugin->u.audit->close != NULL)
+	    debug_return_bool(false);
+    }
+
+    debug_return_bool(true);
+}
+
+/*
+ * Execute a command, potentially in a pty with I/O logging, and
  * wait for it to finish.
  * This is a little bit tricky due to how POSIX job control works and
  * we fact that we have two different controlling terminals to deal with.
  */
 int
-sudo_execute(struct command_details *details, struct command_status *cstat)
+sudo_execute(struct command_details *details,
+    const struct user_details *user_details,
+    struct sudo_event_base *evbase, struct command_status *cstat)
 {
-    debug_decl(sudo_execute, SUDO_DEBUG_EXEC)
+    debug_decl(sudo_execute, SUDO_DEBUG_EXEC);
+
+#if defined(HAVE_SELINUX) && !defined(HAVE_PTRACE_INTERCEPT)
+    /*
+     * SELinux prevents LD_PRELOAD from functioning so we must use
+     * ptrace-based intercept mode.
+     */
+    if (details->selinux_role != NULL || details->selinux_type != NULL) {
+	if (ISSET(details->flags, CD_INTERCEPT)) {
+	    sudo_warnx("%s",
+		U_("intercept mode is not supported with SELinux RBAC on this system"));
+	    CLR(details->flags, CD_INTERCEPT);
+	}
+	if (ISSET(details->flags, CD_LOG_SUBCMDS)) {
+	    sudo_warnx("%s",
+		U_("unable to log sub-commands with SELinux RBAC on this system"));
+	    CLR(details->flags, CD_LOG_SUBCMDS);
+	}
+    }
+#endif /* HAVE_SELINUX && !HAVE_PTRACE_INTERCEPT */
 
     /* If running in background mode, fork and exit. */
     if (ISSET(details->flags, CD_BACKGROUND)) {
@@ -397,36 +463,44 @@ sudo_execute(struct command_details *details, struct command_status *cstat)
 		cstat->val = errno;
 		debug_return_int(-1);
 	    case 0:
-		/* child continues without controlling terminal */
+		/*
+		 * Child continues in an orphaned process group.
+		 * Reads from the terminal fail with EIO.
+		 * Writes succeed unless tostop is set on the terminal.
+		 */
 		(void)setpgid(0, 0);
 		break;
 	    default:
 		/* parent exits (but does not flush buffers) */
 		sudo_debug_exit_int(__func__, __FILE__, __LINE__,
 		    sudo_debug_subsys, 0);
-		_exit(0);
+		_exit(EXIT_SUCCESS);
 	}
     }
 
+    /*
+     * Restore resource limits before running.
+     * We must do this *before* calling the PAM session module.
+     */
+    restore_limits();
+
     /*
      * Run the command in a new pty if there is an I/O plugin or the policy
      * has requested a pty.  If /dev/tty is unavailable and no I/O plugin
      * is configured, this returns false and we run the command without a pty.
      */
     if (sudo_needs_pty(details)) {
-	if (exec_pty(details, cstat))
+	if (exec_pty(details, user_details, evbase, cstat))
 	    goto done;
     }
 
     /*
-     * If we are not running the command in a pty, we were not invoked
-     * as sudoedit, there is no command timeout and there is no close
-     * function, just exec directly.  Only returns on error.
+     * If we are not running the command in a pty, we may be able to
+     * exec directly, depending on the plugins used.
      */
-    if (!ISSET(details->flags, CD_SET_TIMEOUT|CD_SUDOEDIT) &&
-	policy_plugin.u.policy->close == NULL) {
+    if (direct_exec_allowed(details)) {
 	if (!sudo_terminated(cstat)) {
-	    exec_cmnd(details, -1);
+	    exec_cmnd(details, NULL, -1, -1);
 	    cstat->type = CMD_ERRNO;
 	    cstat->val = errno;
 	}
@@ -436,7 +510,7 @@ sudo_execute(struct command_details *details, struct command_status *cstat)
     /*
      * Run the command in the existing tty (if any) and wait for it to finish.
      */
-    exec_nopty(details, cstat);
+    exec_nopty(details, user_details, evbase, cstat);
 
 done:
     /* The caller will run any plugin close functions. */
@@ -478,3 +552,35 @@ terminate_command(pid_t pid, bool use_pgrp)
 
     debug_return;
 }
+
+/*
+ * Free the dynamically-allocated contents of the exec closure.
+ */
+void
+free_exec_closure(struct exec_closure *ec)
+{
+    debug_decl(free_exec_closure, SUDO_DEBUG_EXEC);
+
+    /* Free any remaining intercept resources. */
+    intercept_cleanup(ec);
+
+    sudo_ev_base_free(ec->evbase);
+    sudo_ev_free(ec->backchannel_event);
+    sudo_ev_free(ec->fwdchannel_event);
+    sudo_ev_free(ec->sigint_event);
+    sudo_ev_free(ec->sigquit_event);
+    sudo_ev_free(ec->sigtstp_event);
+    sudo_ev_free(ec->sigterm_event);
+    sudo_ev_free(ec->sighup_event);
+    sudo_ev_free(ec->sigalrm_event);
+    sudo_ev_free(ec->sigpipe_event);
+    sudo_ev_free(ec->sigusr1_event);
+    sudo_ev_free(ec->sigusr2_event);
+    sudo_ev_free(ec->sigchld_event);
+    sudo_ev_free(ec->sigcont_event);
+    sudo_ev_free(ec->siginfo_event);
+    sudo_ev_free(ec->sigwinch_event);
+    free(ec->ptyname);
+
+    debug_return;
+}
